
import sys, os, json, time, requests, csv
from datetime import datetime
import pandas as pd
from collections import ChainMap
from urllib.parse import urlsplit, quote as urlquote

from requests.exceptions import ConnectionError

from eigeningenuity.core.debug import _debug

def _do_eigen_json_request(requesturl,**params):

    if params:
        if "?" in requesturl:
            sep = "&"
        else:
            sep = "?"

        for k,v in params.items():
            if v is None:
               pass
            else:
                for e in force_list(v):
                    if not isinstance(e, str):
                        if isinstance(e, str):
                            e = e.encode("utf8")
                        else:
                            e = str(e)
                    requesturl += sep + urlquote(k) + "=" + urlquote(e.encode("UTF8"))

                    sep = "&"


    _debug("DEBUG",requesturl)
    if 'DEBUG' in os.environ and os.environ['DEBUG']:
        print("DEBUG: [" + requesturl + "]", file=sys.stderr)

    try:
        data = requests.get(requesturl,verify=False)
    except ConnectionError as e:
        # sys.tracebacklimit = 0
        raise EigenException("No Response from ingenuity instance at https://" + requesturl.split("/")[2] + ". Please check this is correct url and that the instance is currently running") from None


    if data.text.startswith("ERROR:"):
        if "UNKNOWN TAG" in data.text:
            raise RemoteServerException(data.text)
        raise RemoteServerException(data.text.split("\n")[1], requesturl)
    elif data.text.startswith("EXCEPT:") and data.status_code != 200:
        raise RemoteServerException( data.text.split("\n")[0], "API could not parse request, check query syntax is correct",)
    else:
        try:
            ret = data.json()
        except ValueError:
            ret = data.text

    return ret

def _do_eigen_post_request(posturl,data):
    requests.post(posturl,data,verify=False)
    pass

def is_list(x):
    return type(x) in (list, tuple, set)

def force_list(x):
    if is_list(x):
        return x
    else:
        return [x]

def number_to_string(n):
    if type(n) == float:
        return format(n, '^12.5f')
    else:
        return n

def time_to_epoch_millis(t):
    if type(t) == datetime:
        epochmillis = time.mktime(t.timetuple()) * 1000
    elif type(t) == tuple:
        epochmillis = time.mktime(t) * 1000
    elif type(t) == int and t > 100000000000:
        epochmillis = t
    elif type(t) == int:
        epochmillis = t*1000
    elif type(t) == float and t > 100000000000:
        epochmillis = int(t)
    elif type(t) == float:
        epochmillis = int(t)*1000
    elif type(t) == str:
        epochmillis = get_timestamp(t)
    else:
        raise EigenException("Unknown time format " + str(type(t)))
    return int(round(epochmillis))

def get_time_tuple(floatingpointepochsecs):
    time_tuple = time.gmtime(floatingpointepochsecs)
    return time_tuple

def get_timestamp_string(t):
    pattern = '%Y-%m-%d %H:%M:%S UTC'
    s = datetime.fromtimestamp(t).strftime(pattern)
    return s

def get_timestamp(t):
    if type(t) == str:
        try:
            pattern = '%Y-%m-%d %H:%M:%S.%f'
            epochmillis = time.mktime(time.strptime(t, pattern))
        except ValueError:
            try:
                pattern = '%Y-%m-%dT%H:%M:%S.%f%z'
                epochmillis = time.mktime(time.strptime(t, pattern))
            except ValueError:
                try:
                    pattern = '%Y-%m-%d %H:%M:%S'
                    epochmillis = time.mktime(time.strptime(t, pattern))
                except ValueError:
                    try:
                        pattern = '%Y-%m-%d'
                        epochmillis = time.mktime(time.strptime(t, pattern))
                    except ValueError:
                        try:
                            epochmillis = int(t)
                        except ValueError:
                            raise EigenException("Unknown time format " + str(type(t)))

    else:
        epochmillis = time_to_epoch_millis(t)
    return epochmillis



def get_datetime(t):
    timestamp = get_timestamp(t)
    return datetime.fromtimestamp(timestamp)

def pythonTimeToServerTime(ts):
# where ts may be supplied as time tuple, datetime or floating point seconds, and server time is (obviously) millis.
    if type(ts) == datetime:
        epochmillis = time.mktime(ts.timetuple()) * 1000
    elif type(ts) == tuple:
        epochmillis = time.mktime(ts) * 1000
    elif type(ts) == float:
        epochmillis = int(ts * 1000)
    else:
        raise EigenException("Unknown python time format " + str(type(ts)))
    return int(round(epochmillis))


def serverTimeToPythonTime(ts):
# where ts is millis and the returned value is consistently whatever we're using internally in the python library (i.e. floating secs)
    return ts / 1000.0

def pythonTimeToFloatingSecs(ts):
# where ts may be supplied as time tuple, datetime or floating point seconds
    if type(ts) == datetime:
        return time.mktime(ts.timetuple())
    elif type(ts) == tuple:
        return time.mktime(ts)
    elif type(ts) == float or type(ts) == int:
        return ts
    else:
        raise EigenException("Unknown python time format " + str(type(ts)))

def pythonTimeToTuple(ts):
# where ts may be supplied as time tuple, datetime or floating point seconds
    if type(ts) == datetime:
        return ts.timetuple()
    elif type(ts) == tuple:
        return ts
    elif type(ts) == float:
        return ts
    else:
        raise EigenException("Unknown python time format " + str(type(ts)))

def pythonTimeToDateTime(ts):
# where ts may be supplied as time tuple, datetime or floating point seconds
    if type(ts) == datetime:
        return ts
    elif type(ts) == tuple:
        epochmillis = time.mktime(ts) * 1000
        return datetime.fromtimestamp(epochmillis)
    elif type(ts) == float:
        return time.gmtime(ts)
    else:
        raise EigenException("Unknown python time format " + str(type(ts)))

def map_java_exception(myExceptionName, params):
    if myExceptionName == "urllib2.HTTPError":
         return RemoteServerException(params)
#    elif: ...
    else:
        return EigenException(myExceptionName)

def parse_duration(timeWindow):
    unit = timeWindow[-1:]
    value = timeWindow[:-1]

    def seconds():
        int(value)

    def minutes():
        return int(value) * 60

    def hours():
        return int(value) * 3600

    def days():
        return int(value) * 3600 * 24

    def months():
        return int(value) * 3600 * 24 * 30

    def years(): int(value) * 3600 * 24 * 365

    options = {"s" : seconds,
               "m" : minutes,
               "h" : hours,
               "d" : days,
               "M" : months,
               "y" : years,
    }

    duration = options[unit]()*1000

    return duration


def divide_chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]

def cypherRespMap(resp):
    return resp["m"]


def jsonToDf(json,transpose:bool=False):
    p = {}
    keys = json.keys()
    if "unknown" in keys: keys.remove("unknown")
    for key in keys:
        if type(json[key]) == list:
            m = list(map(dataMap,json[key]))
            n = dict(ChainMap(*m))
        else:
            n = dataMap(json[key])
        p[key] = n
    try:
        df = pd.DataFrame(p)
        if transpose:
            df = df.T
        return(df[::-1])
    except ValueError:
        return(pd.Series(p))

def dataMap(j):
    try:
        h = {datetime.fromtimestamp(j["timestamp"] / 1000): j["value"]}
    except:
        h = j
    return h

def flattenList(list):
    return [item for sublist in list for item in sublist]

def flattenDict(nestedDict):
    points = []
    keys = nestedDict.keys()
    for key in keys:
        values = nestedDict[key]
        for value in force_list(values):
            tag = {"tag": key}
            tag.update(value)
            points.append(tag)
    return points

def aggMap(x):
    for k in x[1]:
        k["tag"] = x[0]
    return x[1]

def aggToDf(x):
    y = flattenList(list(map(aggMap,x.items())))
    cols = ["tag", "start", "end", "min", "max", "avg", "var", "stddev", "numgood", "numbad"]
    df = pd.DataFrame(y)
    df = df[cols]
    df["start"] = pd.to_datetime(df["start"], unit="ms")
    df["end"] = pd.to_datetime(df["end"], unit="ms")
    df.sort_values(by='start', inplace=True)
    return(df)

def get_eigenserver(url):
    split_url = urlsplit(url)
    return split_url.scheme + "://" + split_url.netloc

def constructURL(y,x):
    x["url"] = y + x["url"]
    k = {"fileName": x["fileName"], "description": x["description"], "url": x["url"]}
    return k

def parseEvents(events):
    if type(events) == dict or type(events) == list:
        events = events
    elif type(events) == str:
        try:
            with open(events,"r") as f:
                events = json.loads(f.read())
        except FileNotFoundError:
            try:
                events = json.loads(events)
            except json.decoder.JSONDecodeError:
                raise EigenException("Could not parse input, enter the path to a file, or a json/dict object")
    try:
        events = events["events"]
    except:
        pass
    events = force_list(events)
    return events

def parse_properties(x):
    return x["graphapi"]["properties"]

def csvWriter(data,historian,filepath,multi_csv,functionName,order=None,headers=False):
    if "Agg" in functionName:
        timeField = 'start'
    else:
        timeField = 'timestamp'
    if multi_csv:
        for item in data:
            sortedDicts=[]
            for entry in force_list(data[item]):
                if order is not None:
                    entry = (dict(sorted(entry.items(), key=lambda item: order.index(item[0]))))
                sortedDicts.append(dict(entry.items()))
            sortedList = sorted(sortedDicts, key=lambda d: (d[timeField])) 
            keys = sortedList[0].keys()
            if filepath is None:
                filepath = item + "-" + functionName + "-" + str(round(datetime.now().timestamp())) + ".csv"
            elif filepath[-1] == "/":
                filepath += item +  "-" + functionName + "-" + str(round(datetime.now().timestamp())) + ".csv"
            else:
                filepath = item +  "-" + functionName + "-" + str(round(datetime.now().timestamp())) + ".csv"

            with open(filepath, 'w', newline='') as output_file:
                dict_writer = csv.DictWriter(output_file, keys)
                if headers:
                    dict_writer.writeheader()
                dict_writer.writerows(sortedList)
        return True       
    else:
        sortedDicts = []
        points = flattenDict(data)
        for entry in points:
            if order is not None:
                entry = dict(sorted(entry.items(), key=lambda item: order.index(item[0])))
            sortedDicts.append(entry)
        sortedList = sorted(sortedDicts, key=lambda d: (d['tag'], d[timeField])) 
        keys = sortedList[0].keys()
        if filepath is None:
            filepath = historian +  "-" + functionName + "-" + str(round(datetime.now().timestamp())) + ".csv"
        elif filepath[-1] == "/":
            filepath += historian +  "-" + functionName + "-" + str(round(datetime.now().timestamp())) + ".csv"

        with open(filepath, 'w', newline='') as output_file:
            dict_writer = csv.DictWriter(output_file, keys)
            if headers:
                dict_writer.writeheader()
            dict_writer.writerows(sortedList)
            return True



class EigenException (Exception): pass
class RemoteServerException (EigenException): pass



"""Eigen Ingenuity - Asset Model

This package deals with the Eigen Ingenuity Asset Model API, by means
of the Cypher plugin endpoint.

To retrieve an AssetObject use getMatchingNodes, or execute a custom cypher query with executeRawQuery

  from eigeningenuity.assetmodel import getAssetModel
  from eigeningenuity import EigenServer

  eigenserver = EigenServer(ingenuity-base-url)

  model = getAssetModel(eigenserver)

  nodes = model.getMatchingNodes("code","System_")
  
    
"""
import functools

import requests, os
import json
import pandas as pd
from eigeningenuity import EigenServer
from urllib.parse import quote as urlquote
from eigeningenuity.util import force_list, _do_eigen_json_request, parse_properties, get_eigenserver, constructURL, cypherRespMap
from eigeningenuity.core import get_default_server
from requests.exceptions import ConnectionError
from urllib.error import URLError
from typing import Union
import datetime as dt
from datetime import datetime

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

class AssetModel (object):
    """An assetmodel instance which talks the Eigen Neo4j endpoint.
    """
    def __init__(self, baseurl):
        """This is a constructor. It takes in a URL like https://demo.eigen.co/ei-applet/"""
        self.baseurl = baseurl
        self.eigenserver = get_eigenserver(baseurl)

    def _testConnection(self):
        """Preflight Request to verify connection to ingenuity"""
        try:
            status = requests.get(self.baseurl).status_code
        except (URLError, ConnectionError):
            raise ConnectionError("Failed to connect to ingenuity instance at " + self.baseurl + ". Please check the url is correct and the instance is up.")

    def _doJsonCypherRequest(self, cmd, params):
        self._testConnection()
        url = self.baseurl + "jcypher?cmd=" + urlquote(cmd)
        return _do_eigen_json_request(url, **params)

    def _doJsonCommonMenuRequest(self, index, params):
        self._testConnection()
        url = self.baseurl + "commonmenu2/asset/" + index + "?"
        return _do_eigen_json_request(url, **params)

    def executeRawQuery(self,query:str,output:str="json"):
        """
        Executes a raw cypher query against AssetModel via the AssetModel API.

        Args: 
            query: A string containing a query in cypher, for information on cypher queries see https://neo4j.com/developer/cypher/

        Returns:
            A Json containing the cypher response to the query.

        Raises:
            KeyError: Raises an exception
            
        """
        args = {'q': query}
        response = self._doJsonCypherRequest("execute", args)

        if output == "raw":
            return response
        if output == "json":
            return response
        if output == "df":
            return pd.json_normalize(response)



    def getRelatedAssetsCommonMenu(self, node: str, output="json", filepath = None):
        """
        Return all measurement tags directly related to a given asset via the Common Menu API.

        Args:
            node:
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            A Json containing all nodes with a relation to any node that meets the criteria, and their relationship type.

        Raises:
            KeyError: Raises an exception

        """
        args = {"asset": node}

        response = self._doJsonCommonMenuRequest("relatedAssets", args)
        relatedassets = response["relatedAssets"]["graphapi"]

        if output == "raw":
            return response
        if output == "json":
            return relatedassets
        if output == "df":
            return pd.json_normalize(relatedassets).drop(columns=["asset.externalId"])
        if output == "file":
            if filepath is None:
                filepath = node + "-RelatedAssets-" + str(dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
            with open(filepath + ".json", "w") as f:
                f.write(json.dumps(relatedassets, indent=4))

    def getRelatedMeasurementsCypher(self, query:str, prop:str="code", measurement:str="", output="json"):
        """
        Return all measurement tags directly related to a given asset via the AssetModel API.

        Args: 
            nodes: The value of the property to match
            prop: (Optional) The property on to query nodes on. Defaults to code.
            measurement: (Optional) Return only tags for a specified measurement type. Defaults to return all tags.

        Returns:
            A Json containing all nodes with a relation to any node that meets the criteria, and their relationship type.

        Raises:
            KeyError: Raises an exception
            
        """
        result = []
        exact = True
        for item in force_list(query):
            x = []
            y = []
            args = {}
            # if exact:
            matchStatement = "n.%s = '%s' and r.measurementName contains '%s'" % (prop, item, measurement)
            # else:
            #     matchStatement = "n.%s contains '%s' and r.measurementName contains '%s'" % (prop, node, measurement)
            args['q'] = 'MATCH (n)-[r:hashistoriantag]->(m) WHERE %s RETURN r.measurementName as measurementName, m.code as measurementCode' % matchStatement

            response = self._doJsonCypherRequest("execute", args)


            for meas in response:
                x.append(meas["measurementName"])
                y.append(meas["measurementCode"])

            result.append(pd.Series(y,index=x,name=item))
        k = pd.merge(result[0],result[1],on="index")
        if output == "raw":
            return response
        if output == "json":
            return response
        if output == "df":
            return pd.DataFrame(response)

    def getRelatedMeasurementsCommonMenu(self, node:str, output="json"):
        """
        Return all measurement tags directly related to a given asset via the Common Menu API.

        Args: 
            node: The code (ID) of the node to get related measurements from
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            A Json containing all nodes with a relation to any node that meets the criteria, and their relationship type.

        Raises:
            KeyError: Raises an exception
            
        """
        args = {}
        args["asset"] = node

        response = self._doJsonCommonMenuRequest("timeseries", args)
        timeseries = response["timeseries"]

        if output == "raw":
            return response
        if output == "json":
            return timeseries
        if output == "df":
            return pd.DataFrame(timeseries)["tag"]

    def getRelatedDocuments(self, node:Union[str,list], match=None, output="json", directory:str=None):
        """
        Return all Documents related to nodes via the Common Menu API.

        Args:
            node: The name of the node to query documents for
            match: (Optional) Filter returned Documents to those with filenames matching a string
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API; "json" - A processed version of the json response; "df" - A formatted pandas dataframe object; "download" - Downloads the files to a local directory"
            directory: (Optional) Name and path to the directory created/used for downloaded documents. If omitted, will download files to current directory. Has no effect unless output is "download".

        Returns:
            The set of all documents related to the nodes, the format is dependent on the output parameter. If output is "download", it returns instead the files themselves

        Raises:
            KeyError: Raises an exception

        """
        nodes = force_list(node)
        results = []
        documents = []
        for node in nodes:
            args = {}
            args["asset"] = node
            response = self._doJsonCommonMenuRequest("documents", args)
            results.append(response)
            documents.append(response["documents"])

        if documents == [[]]:
            return []
        documents = documents[0][0]["documents"]

        documents = list(map(functools.partial(constructURL, self.eigenserver), documents))

        matches = []
        if match is not None:
            for document in documents:
                if match in document['fileName'] or match in document['description']:
                    matches.append(document)
            documents = matches

        if output == "raw":
            return results
        if output == "json":
            return documents
        if output == "df":
            return pd.json_normalize(documents)
        if output == "download":
            if directory is None:
                directory = "."
            os.makedirs(directory, exist_ok=True)
            for doc in documents:
                with requests.get(doc["url"], stream=True,timeout=1000) as r:
                    r.raise_for_status()
                    with open(directory + "/" + doc["fileName"], 'wb') as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)
                        f.close()


    def getLabels(self):
        """
        List all labels used in an assetmodel instance

        Args:
            N/A
        Returns:
            A list of all node labels/types present in the model
        """
        args = {"q": "call db.labels()"}
        response = self._doJsonCypherRequest("EXECUTE",args)

        labels = []
        for label in response:
            labels.append(label["label"])

        return labels

    def getRelatedEvents(self, node:Union[str,list], start:Union[str,int,float,datetime]="24 hours ago", end:Union[str,int,float,datetime]="now", limit:int=1000, output:str="json", directory:str=None):
        """
        Return all Events related to nodes via the Common Menu API.

        Args:
            node: The name of the node to query documents for
            start: Timestamp of beginning of time window to search for events, also accepts strings like "30 mins ago"
            end: Timestamp of end of time window to search for events
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API; "json" - A processed version of the json response; "df" - A formatted pandas dataframe object; "download" - Downloads the files to a local directory"
            directory: (Optional) Name and path to the directory created/used for downloaded documents. If omitted, will download files to current directory. Has no effect unless output is "download".

        Returns:
            The set of all documents related to the nodes, the format is dependent on the output parameter. If output is "download", it returns instead the files themselves

        Raises:
            KeyError: Raises an exception

        """
        nodes = force_list(node)
        results = []
        events = []
        for node in nodes:
            args = {"asset": node, "start": start, "end": end, "limit": limit}
            response = self._doJsonCommonMenuRequest("events", args)
            results.append(response)
            events.append(response["events"])

        if events == []:
            return 


        if output == "raw":
            return results
        if output == "json":
            return events[0]
        if output == "df":
            return pd.json_normalize(events[0])

    def getProperties(self, nodes:Union[str,list], output="json"):
        """
        Return all properties of a node via the Assetmodel API

        Args:
            node: The name of the node to return properties of
            output:

        Returns:
            A Json containing all nodes with a relation to any node that meets the criteria, and their relationship type.

        Raises:
            KeyError: Raises an exception

        """
        nodes = force_list(nodes)
        results = []
        properties = []
        for node in nodes:
            args = {}
            args["asset"] = node
            response = self._doJsonCommonMenuRequest("properties", args)
            results.append(response)
            properties.append(response["properties"])

        if output == "raw":
            return results
        if output == "json":
            return properties
        if output == "df":
            data = list(map(parse_properties, properties))
            return pd.DataFrame(data,nodes)


    def getMatchingNodes(self,prop:str,node:str,exact:bool=False):
        """
        Return all properties of matching nodes via a cypher query response
    
        Args:
            prop: The property on which to match
            match: A list of nodes to return properties for
            exact: (Optional) Whether to only return nodes where the prop is equal to match, or nodes where prop contains match. Defaults to False
    
        Returns:
            A Json containing all nodes that contains the criteria.
    
        Raises:
            KeyError: Raises an exception
        """
    
        args = []
        if exact:
            matchStatement = "n.%s = '%s'" % (prop,node)
        else:
            matchStatement = "n.%s contains '%s'" % (prop,node)
        args['q'] = 'MATCH (n) WHERE %s RETURN n' % matchStatement
    
        return self._doJsonCypherRequest("execute", args)
    
    def getRelatedAssetsCypher(self, prop, nodes, exact: bool = False, relation=None):
        """
        Return all assets directly related to a given asset.
    
        Args:
            prop: The property on which to match a node
            nodes: The value of the property to match
            exact: Whether or not to only to apply to assets exactly matching value of node (Defaults to False)
            relation: Specify a required relationship type for returned nodes
    
        Returns:
            A Json containing all nodes with a relation to any node that meets the criteria, and their relationship type.
    
        Raises:
            KeyError: Raises an exception
    
        """
        ret = []
        exact = True
        for node in force_list(nodes):
            args = {}
            if relation is not None: relation = ":" + relation
            if exact:
                matchStatement = "n.%s = '%s'" % (prop, node)
            else:
                matchStatement = "n.%s contains '%s'" % (prop, node)
            args['q'] = 'MATCH (n)-[r%s]-(m:AssetObject) WHERE %s RETURN m' % (relation, matchStatement)
    
        return list(map(cypherRespMap, self._doJsonCypherRequest("execute", args)))  # Func was map of k -> k["m"]

def get_assetmodel(eigenserver:EigenServer=None):
    """
    Connect to Assetmodel of eigenserver. If eigenserver is not provided this will default to the EIGENSERVER environmental variable
    
    Args:
        eigenserver: An instance of EigenServer() to query

    Returns:
        An object defining a connection to the AssetModel
    """
    if eigenserver is None:
            eigenserver = get_default_server()

    return AssetModel(eigenserver.getEigenServerUrl() + "ei-applet" + "/")



"""Eigen Ingenuity - SQL

This package deals with the Eigen Ingenuity SQL database.

To retrieve data execute a custom query with

  from eigeningenuity.assetmodel import matchNode
  from time import gmtime, asctime

  h = get_historian("Demo-influxdb","demo.eigen.co")

  cmd = "EXECUTE"
  db = "aveva-db",
  query = '''
    SELECT TAG_TISLOCATIONCODE, TAG_TISLOCATIONNAME, TAG_NUMBER, TAG_DESC, TAG_STATUS, TAG_STATUSDESC, TAG_BARRIERCODE, TAG_BARRIERNAME, TAG_BARRIERISACTIVE, TAG_BARRIERPERFORMANCESTD, TAG_BARRIERVERIFICATIONACT
    FROM VT_TAGBARRIER
    WHERE (ROWNUM <= 10)
    '''

  response = h.executeRawQuery(cmd,db,query)
"""

from eigeningenuity import EigenServer
from urllib.parse import quote as urlquote
from eigeningenuity.util import _do_eigen_json_request
from eigeningenuity.core import get_default_server
import requests, json
import pandas as pd
import datetime as dt

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

class sqlConnection (object):
    """An elasticsearch instance which talks the Eigen elastic endpoint.
    """
    def __init__(self, baseurl):
       """This is a constructor. It takes in a URL like http://demo.eigen.co/eigen-webtools-servlet/sqlquery"""
       self.baseurl = baseurl

    def _testConnection(self):
        """Preflight Request to verify connection"""
        try:
            status = requests.get(self.baseurl).status_code
        except ConnectionError:
            raise ConnectionError("Failed to connect to ingenuity instance at " + self.baseurl + ". Please check the url is correct and the instance is up.")

    def _doJsonRequest(self,cmd,params):
        url = self.baseurl + "?cmd=" + urlquote(cmd)
        return _do_eigen_json_request(url, **params)

    def executeRawQuery(self,cmd:str,db:str,query:str,output:str="json",filepath:str=None):
        """
        Run a user composed raw SQL query against an ingenuity SQL database

        Args:
            cmd: Command to run against the database e.g. EXECUTE
            db: Name of the SQL database to query
            query: The Body of the SQL query
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API; "json" - A processed version of the json response; "df" - A formatted pandas dataframe object; "file" - Writes the response to a .json file in the local directory. Defaults to "json"
            filepath: (Optional) Name and path to the .json file that will be created/overwritten. If omitted, will create a file in the current directory with an auto-generated name. Has no effect unless output is "file".

        Returns:
            The SQL response to the given query, the format is dependent on the output parameter
        """
        args = {}
        args['db'] = db
        args['query'] = query
        response = self._doJsonRequest(cmd,args)["results"]

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(response)
        elif output == "df":
            return pd.DataFrame(response)
        elif output == "file":
            if filepath is None:
                filepath = "eigenSQLResponse-" + db + "-" + str(dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
            with open(filepath + ".json", "w") as f:
                f.write(json.dumps(response, indent=4))
        return None


def get_sql(eigenserver:EigenServer = None):
    """Instantiate an SQL connection for the given EigenServer.

    Args:
        eigenserver: (Optional) An EigenServer Object linked to the ingenuity url containing the elasticsearch. Can be omitted if environmental variable "EIGENSERVER" exists and contains the Ingenuity base url

    Returns:
        An Object that can be used to query SQL databases connected to the ingenuity.
    """
    if eigenserver is None:
        eigenserver = get_default_server()
    return sqlConnection(eigenserver.getEigenServerUrl() + "eigen-webtools-servlet/sqlquery")



from __future__ import (absolute_import, division, print_function, unicode_literals)

import sys, os

debuglevels = {"CRITICAL": 1, "ERROR": 2, "WARNING":3, "INFO":4, "DEBUG":5, "TRACE":6}

def _debug(debuglevelstr, message):
    """Checks for current level of debug requested, and prints message on condition that the level is met or exceeded"""
    userrequestlevelstr = os.getenv("DEBUG")
    if userrequestlevelstr:  
        if (userrequestlevelstr == "0") or (userrequestlevelstr == "no"):
            userrequestlevelstr = "ERROR"
        elif (userrequestlevelstr == "1") or (userrequestlevelstr == "yes"):
            userrequestlevelstr = "DEBUG"
    else:
        userrequestlevelstr = "ERROR" 
    try:
        requestedlevel = debuglevels.get(userrequestlevelstr)
    except KeyError:
        raise EigenException("Unable to recognise this DEBUG option " + userrequestlevelstr)

    try:
        debuglevel = debuglevels.get(debuglevelstr)
    except KeyError:
        raise EigenException("Unable to recognise this DEBUG option " + userrequestlevelstr)

    if requestedlevel > debuglevel:
        sys.stderr.write(message + "\n")



"""Eigen Ingenuity - Elasticsearch

This package deals with the Eigen Ingenuity Elasticsearch database.

To retrieve an AssetObject use matchNode, or execute a custom cypher query with 

  from eigeningenuity.assetmodel import matchNode
  from time import gmtime, asctime

  nodes = matchNodes("code","System_")
  
  for node in nodes:
      code = node.code
      print(code)  
"""

from eigeningenuity import EigenServer
from urllib.parse import quote as urlquote
from eigeningenuity.historian import get_historian
from eigeningenuity.util import is_list,force_list,_do_eigen_json_request,EigenException
from eigeningenuity.core import get_default_server
import pandas as pd
import datetime as dt
import json
import requests
from requests.exceptions import ConnectionError
from urllib.error import URLError

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

class ElasticConnection (object):
    """An elasticsearch instance which talks the Eigen elastic endpoint.
    """

    def __init__(self, baseurl):
       """This is a constructor. It takes in a URL like http://infra:8080/ei-applet/search/"""
       self.baseurl = baseurl

    def _doJsonRequest(self,cmd,params):
        url = self.baseurl + "?cmd=" + urlquote(cmd)
        return _do_eigen_json_request(url,**params)

    def _testConnection(self,instance):
        """Preflight Request to verify connection to ingenuity"""
        try:
            status = requests.get(self.baseurl + "?cmd=DODIRECTSEARCH&clientname=" + instance).status_code
            if status != 200:
                raise ConnectionError(
                    "Invalid API Response from " + self.baseurl + "?cmd=DODIRECTSEARCH&clientname" + instance + ". Please check the base url is correct, the instance is running and has an elasticsearch database.")
        except (URLError, ConnectionError):
            raise ConnectionError ("Failed to connect to ingenuity instance at " + self.baseurl + "?cmd=DODIRECTSEARCH" + ". Please check the url is correct and the instance is up.")

    def executeRawQuery(self,index:str, query:str, instance:str = "elasticsearch-int", output:str = "json", filepath:str = None):
        """Executes a raw cypher query against Elastic.

        Args:
            index: The elasticsearch index to query
            query: The body of the query
            instance: (Optional) The instance of elasticsearch to query. Defaults to elasticsearch-int
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API; "json" - A processed version of the json response; "df" - A formatted pandas dataframe object; "file" - Writes the response to a .json file in the local directory. Defaults to "json"
            filepath: (Optional) Name and path to the .json file that will be created/overwritten. If omitted, will create a file in the current directory with a generated name. Has no effect unless output is "file".
            
        Returns:
            Elasticsearch API response to the query, the format is dependent on the output parameter
        """
        self._testConnection()
        validOutputTypes = ["raw", "json", "df", "file"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args["clientname"] = instance
        args["index"] = index
        args["search"] = query

        response = self._doJsonRequest("DODIRECTSEARCH", args)

        if "error" in response.keys():
            raise EigenException("Invalid request", response["error"])

        if output == "raw":
            return response
        elif output == "json":
            return response['results']
        elif output == "df":
            return pd.json_normalize(response['results'])
        elif output == "file":
            if filepath is None:
                filepath = "eigenElasticResponse-" + index + "-" + str(dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
            with open(filepath + ".json", "w") as f:
                f.write(json.dumps(response, indent=4))


def get_elastic(eigenserver:EigenServer = None):
    """Instantiate an elasticsearch connection for the given EigenServer.

    Args:
        eigenserver: (Optional) An EigenServer Object linked to the ingenuity url containing the elasticsearch. Can be omitted if environmental variable "EIGENSERVER" exists and is equal to the Ingenuity base url

    Returns:
        An Object that can be used to query elasticsearch data from the ingenuity.
    """
    if eigenserver is None:
        eigenserver = get_default_server()
    return ElasticConnection(eigenserver.getEigenServerUrl() + "ei-applet/search")


"""Eigen Ingenuity - Smart Dash

This package deals with the Eigen Ingenuity Smart Dash API.

To get a smart object to work with, use get_historian(xxx) with either
an instance name (which will be resolved via the the usual methods in
eigeningenuity.core) or a full URL to a JSON Bridge Historian instance.

  from eigeningenuity.smartdash import get_smartdash

  sd = get_smartdash()
  sd.addItem("foobar", "HistorianTagTotaliser", "pi-af/testsim00001", "year")
  vals = sd.getValues()

  print vals['foobar']

 """

from __future__ import (absolute_import, division, print_function, unicode_literals)

import requests

from eigeningenuity.core import get_default_server
from eigeningenuity.util import _do_eigen_json_request, force_list, time_to_epoch_millis, is_list, get_datetime, number_to_string, EigenException, get_timestamp_string, pythonTimeToFloatingSecs, serverTimeToPythonTime, pythonTimeToServerTime, get_time_tuple
from urllib.parse import quote_plus as urlquote_plus



class SmartDashSet (object):
    def __init__(self, baseurl):
        self.__items = {}
        self.__key = None
        self.__baseurl = baseurl
        self.__itemcount = 0

    def addAutoItem(self, itype, *args):
        self.__itemcount += 1
        id = "item%04d" % self.__itemcount
        self.addItem(id, itype, *args)

        return id
        
    def clear(self):
        self.__key = None
        self.__items = {}
        
    def removeItem(self, kname):
        self.__key = None
        del self.__items[kname]

    def addItem(self, kname, itype, *args):
        self.__key = None
        encodedargs = []
        for i in ((itype,) + args):
            encodedargs.append(urlquote_plus(str(i)))

        itemSpec = ":".join(encodedargs)
        self.__items[urlquote_plus(kname)] = itemSpec

    def getKey(self):
        if self.__key is None:
            self.getValues()
        return self.__key

    def getValues(self):
        res = None

        if self.__key is not None:
            try:
                res = _do_eigen_json_request(self.__baseurl, key = self.__key)
            except EigenException:
                res = None

        if res is None:
            items = []
            for k,spec in self.__items.items():
                items.append(k + "=" + spec)

            res = _do_eigen_json_request(self.__baseurl, item = items)

        self.__key = res['key']
        return res['data']

def get_smartdash(eigenserver = None):
    if eigenserver is None:
        eigenserver = get_default_server()

    url = eigenserver.getEigenServerUrl() + "smartdash-servlet"

    x = requests.get(url).status_code

    if x != 200:
        url = eigenserver.getAppUrl("dashboards-applet")

    if url is None:
        raise EigenException("No smartdash found")

    return SmartDashSet(url + "/smartdash")





import requests

from eigeningenuity.core import get_default_server, EigenServer
from eigeningenuity.util import get_eigenserver, divide_chunks, parseEvents

from typing import Union

class EventLog (object):
    """An elasticsearch instance which talks the Eigen elastic endpoint.
    """
    def __init__(self, baseurl):
        """This is a constructor. It takes in a URL like http://infra:8080/ei-applet/search/"""
        self.baseurl = baseurl
        self.eigenserver = get_eigenserver(baseurl)

    def _testConnection(self):
        """Preflight Request to verify connection to ingenuity"""
        try:
            status = requests.get(self.baseurl).status_code
        except (ConnectionError):
            raise ConnectionError ("Failed to connect to ingenuity instance at " + self.eigenserver + ". Please check the url is correct and the instance is up.")

    def pushToEventlog(self, events):
        """
        Push one or more events to the ingenuity eventlog, accepts any event structure

        Args:
            events: A single event as dict, many events as a list of dicts, or the string filepath of a file containing events

        Returns:
            A boolean representing the successful push of all events. False if at least one event failed to be created
        """
        events = parseEvents(events)
        url = self.baseurl + "events/save-multiple"
        event_chunks = list(divide_chunks(events, 500))
        success = []
        for chunk in event_chunks:
            data = {"events": chunk}
            success.append(requests.post(url, json=data, verify=False))
        return all(success)

    def pushTo365Events(self, events:Union[dict,list,str]) -> bool:
        """
        Push one or more events to the office 365 connector, only accepts pre-defined event structures

        Args:
            events: Can only accept event structures that have been pre-defined in ingenuity. A single event as dict, many events as a list of dicts, or the string filepath of a file containing events

        Returns:
            A boolean representing the successful push of all events. False if at least one event failed to be created
        """
        events = parseEvents(events)
        url = self.baseurl + "eventbus/publish-multiple"
        event_chunks = list(divide_chunks(events, 500))
        success = []
        for chunk in event_chunks:
            data = {"events": chunk}
            success.append(requests.post(url, json=data, verify=False))
        return all(success)

def get_eventlog(eigenserver:EigenServer=None):
    """
    Connect to Assetmodel of eigenserver. If eigenserver is not provided this will default to the EIGENSERVER environmental variable

    Args:
        eigenserver: An instance of EigenServer() to query

    Returns:
        An object defining a connection to the AssetModel
    """
    if eigenserver is None:
            eigenserver = get_default_server()

    return EventLog(eigenserver.getEigenServerUrl() + "eventlog-servlet" + "/")




#!/usr/bin/env python

from setuptools import setup,find_packages
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

pkgname = 'python-eigen-ingenuity'

# Invoke setup
setup(
    name=pkgname,
    version='0.4.2',
    author='Murray Callander',
    author_email='info@eigen.co',
    url='https://www.eigen.co/',
    description="A python library used to query data from the Eigen Ingenuity system",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages("."),
    license='Apache License 2.0',
    install_requires=["pandas",
                    "requests",
                    "urllib3"],
)



"""Eigen Ingenuity

Interface to the Eigen Ingenuity system.

Set the EIGENSERVER environment variable either to hostname, hostname:port
or http://hostname:port/prefix/path as appropriate. It defaults to
localhost:8080.

The various get_XXX methods instantiate connections to different parts
of the Eigen Ingenuity infrastructure.

Currently the only supported subsystem is historian.

h = get_historian("instancename")
h.listDataTags()
"""

from __future__ import (absolute_import, division, print_function, unicode_literals)
import pkg_resources
from eigeningenuity.historian import get_historian, list_historians, get_default_historian_name

from eigeningenuity.core import *

__all__ = ["get_historian", "get_assetmodel", "get_elastic", "getsql", "get_eventlog", "list_historians", "get_default_historian_name", "EigenServer"]
__version__ = pkg_resources.require("python-eigen-ingenuity")[0].version



"""Eigen Ingenuity - Core Functions

This module deals with basic common tasks to do with interacting with an
Eigen Ingenuity server.

In order to find a particular subsystem ("app"), you can use e.g.

   ei = EigenServer(servername)
   ei.get_historian(datasource)

If no server name or URL is given it searches sys.argv for a --eigenserver
argument, falling back on the environment variable EIGENSERVER.
This may either be in the form of a URL (http://foo:8087/) or a
hostname, or hostname:port combination. If no variable is set, it defaults
to http://localhost:8080/
"""

import sys, os
from eigeningenuity.util import _do_eigen_json_request, EigenException, RemoteServerException, serverTimeToPythonTime

_INSTANCE = None
class EigenServer (object):
    def __init__(self, name:str = None, disableSsl = False):
        """Takes the base URL for the Eigen Ingenuity server.
        """
        if name is None:
            try:
                for i in range(len(sys.argv)):
                    arg = sys.argv[i]
    
                    if arg.startswith("--eigenserver="):
                        name = arg[14 : ]
                        break
                    elif arg == "--eigenserver":
                        name = sys.argv[i + 1]
                        break
                else:
                    name = os.environ['EIGENSERVER']
            except:
                pass
        
        try:
            if name.startswith("http://") or name.startswith("https://"):
                if name.endswith("/"):
                    baseurl = name
                else:
                    baseurl = name + "/"
            else:
                baseurl = "http://" + name + "/"
        except:
            baseurl = "http://localhost:8080/"

        self.__baseurl = baseurl
        self.__disablessl = disableSsl


    def getEigenServerUrl(self):
        """Find the base URL for the Eigen Ingenuity server.
        """
        return self.__baseurl
    
    def getAppsInfo(self):
        """Request information about install Eigen Ingenuity apps/components.
        """
        return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", cmd = "GETAPPS", _cachetime = 10)
    
    def listApps(self):
        return list(self.getAppsInfo().keys())
    
    def getAppUrl(self, appname):
        """Find the base URL for a specified Eigen Ingenuity apps. Note this is
        something like "historian-servlet" not "trend" or "jsonbridgehistorian".
        Think "module" rather than "tool".
        Returns None if the app is not installed.
        """
        apps = self.getAppsInfo()
        if appname in apps:
           # Remove leading / from appname if present
           appurl = apps[appname]['url']
           if appurl.startswith('/'):
              appurl = appurl[1:]
           print(self.__baseurl + appurl)
           return self.__baseurl + appurl
        else:
           return None
    
    def listDataSourceTypes(self):
        """Request information about installed Eigen Ingenuity Factory Types.
        """
        return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", cmd = "LISTFACTORYTYPES", _cachetime = 10)
    
    def listDataSources(self, type):
        """List the instances for a specified Eigen Ingenuity Factory Type
        Returns None if there are no instances
        """
        data_source_types = self.listDataSourceTypes()
        if type in data_source_types:
           args = {}
           args['cmd'] = "LISTFACTORYINSTANCES"
           args['type'] = type
           return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", **args)
        else:
           return None
    
    def getDefaultDataSource(self, type):
        """Gets the default instance for the named Factory Type. Returns None if none has been specified."""
        data_source_types = self.listDataSourceTypes()
        if type in data_source_types:
            args = {}
            args['cmd'] = "GETDEFAULTFACTORYINSTANCE"
            args['type'] = type
            try: 
                return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", **args)
            except RemoteServerException:
                return None
        else:
            return None
    
    def listFeatureProperties(self):
        """Request information about installed Eigen Ingenuity Feature Properties.
        """
        return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", cmd = "LISTFEATUREPROPERTIES")
    
    def getFeatureProperty(self, property):
        """Get the value of the specified Eigen Ingenuity Feature Property
        """
        args = {}
        args['cmd'] = "GETFEATUREPROPERTY"
        args['property'] = property
        try:
            return  _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", **args)
        except RemoteServerException:
            return None
    
    def getServerTimeInFloatingPointMillis(self):
        """Get the current time from the Server in floating point milliseconds.
        """
        args = {}
        args['cmd'] = "GETSERVERTIME"
        try:
            ts = _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", **args)
            return serverTimeToPythonTime(ts)
        except RemoteServerException:
            return None
    
    def getServerTimezone(self):
        """Get the current timezone and DST status from the Server
        """
        args = {}
        args['cmd'] = "GETSERVERTIMEZONE"
        try:
            tz = _do_eigen_json_request(self.__baseurl + "eigenpluscore-servlet/prodseminfo", **args)
            return tz
        except RemoteServerException:
            return None

    def getHistorian(self, *args, **kwargs):
        from eigeningenuity.historian import get_historian
        return get_historian(*args, eigenserver=self, **kwargs)
    
    def getAssetModel(self, *args, **kwargs):
        from eigeningenuity.assetmodel import get_assetmodel
        return get_assetmodel(*args, eigenserver=self, **kwargs)
    
    def getSmartdash(self, *args, **kwargs):
        from eigeningenuity.smartdash import get_smartdash
        return get_smartdash(*args, eigenserver=self, **kwargs)
    
    def getDebugletsForApp(self, *args, **kwargs):
        from eigeningenuity.debuglets import get_debuglets_for_app
        return get_debuglets_for_app(*args, eigenserver=self, **kwargs)
    
    def getDebuglets(self, *args, **kwargs):
        from eigeningenuity.debuglets import get_debuglets
        return get_debuglets(*args, eigenserver=self, **kwargs)
    
def get_default_server():
    global _INSTANCE

    if _INSTANCE is None:
        _INSTANCE = EigenServer()

    return _INSTANCE

def get_eigenserver_url():
    return get_default_server().getEigenserverUrl()

def get_apps_info():
    return get_default_server().getAppsInfo()
    
def list_apps():
    return get_default_server().listApps()

def get_app_url(appname):
    return get_default_server().getAppUrl(appname)

def list_data_source_types():
    return get_default_server().listDataSourceTypes()

def list_data_sources(type):
    return get_default_server().listDataSources(type)
    
def get_default_data_source(type):
    return get_default_server().getDefaultDataSource(type)

def list_feature_properties():
    return get_default_server().listFeatureProperties()

def get_feature_property(property):
    return get_default_server().getFeatureProperty(property)

def get_server_time_in_floating_point_millis():
    return get_default_server().getServerTimeInFloatingPointMillis()

def get_server_timezone():
    return get_default_server().getServerTimezone()



"""Eigen Ingenuity - Historian

This package deals with the Eigen Ingenuity Historian API, mostly by means
of the JSON Bridge Historian.

To get a historian object to work with, use get_historian(xxx) with either
an instance name (which will be resolved via the the usual methods in
eigeningenuity.core) or a full URL to a JSON Bridge Historian instance.

  from eigeningenuity.historian import get_historian
  from time import gmtime, asctime

  h = get_historian("pi-opc")
  tags = h.listDataTags()
  for tag in tags:
      dp = h.getCurrentDataPoint(tag)
      print(asctime(time.gmtime(dp['timestamp'])), dp['value'])
"""

import json
import datetime
import math
import logging
import csv
import requests
from urllib.parse import quote as urlquote
from typing import Union
from datetime import datetime

from requests.exceptions import ConnectionError
from urllib.error import URLError
from eigeningenuity.core import get_default_server, EigenServer
from eigeningenuity.util import _do_eigen_json_request, force_list, time_to_epoch_millis, is_list, \
    get_datetime, number_to_string, EigenException, get_timestamp_string, pythonTimeToFloatingSecs, \
    serverTimeToPythonTime, pythonTimeToServerTime, get_time_tuple, parse_duration, jsonToDf, \
    aggToDf, divide_chunks, flattenDict, csvWriter

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

logging.basicConfig(format='%(levelname)s - %(message)s', level=logging.INFO)


class JsonBridgeHistorian (object):
    """A Historian which talks the Eigen Historian Json Bridge protocol.
    """

    def __init__(self, serverurl):
        """This is a constructor. It takes in a URL like http://infra:8080/historian-servlet/jsonbridge/Demo-influxdb"""
        self.serverurl = serverurl
        self.historian = serverurl.split("/")[-1]
        self.baseurl = serverurl.rsplit("/",1)[0]

    def _testConnection(self):
        try:
            status = requests.get(self.serverurl).status_code
            if status != 200:
                raise ConnectionError(
                    "Invalid API Response from " + self.serverurl + ". Please check the url is correct and the instance is up.")
        except (URLError, ConnectionError):
            raise ConnectionError ("Failed to connect to ingenuity instance at " + self.baseurl + ". Please check the url is correct and the instance is up.")

    def listDataTags(self, wildcard):
        """Return a list of all tags that match a given wildcard, or all tags in the historian

        Args:
            wildcard: (Optional) Return all tags that contain this value in their name. Defaults to All Tags in historian.

        Returns:
            A list of all tags containing wildcard

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        self._testConnection()
        args = {}
        if wildcard is not None:
            args['match'] = force_list(wildcard)

        return self._doJsonRequest("list", args)

    def getMetaData(self, tags: Union[str, list], output: str = "json", filepath:str=None, multi_csv:bool=False):
        """
        Return a specified number of interpolated datapoints between a start and end timestamp for one or more datatags

        Args:
            tags: A datatag, or list of datatags
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Units, unit multiplier, and description for each datatag, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        ret = {}

        for tag in force_list(tags):
            args = {}
            args['tag'] = tag
            ret[tag] = (self._doJsonRequest("getmeta", args))

        if output == "raw":
            return ret
        if output == "json":
            return json.dumps(ret)
        if output == "df":
            return jsonToDf(ret, True)
        elif output == "csv":
            return csvWriter(ret,self.historian,filepath,multi_csv,"GetMetaData")
        return None


    def getCurrentDataPoints(self, tags: Union[str, list], output: str = "json", filepath:str=None, multi_csv:bool=False):
        """
        Return most recent raw value for one or more datatags

        Args:
            tags: A datatag, or list of datatags
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            A timestamps, values and statuses for each tag, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args['tag'] = force_list(tags)
        response = self._doJsonRequest("getmulticurrent", args)
        items = response["items"]

        if response['unknown'] != []:
            if items == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(response["unknown"]) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(items)
        elif output == "df":
            return jsonToDf(items)
        elif output == "csv":
            return csvWriter(items,self.historian,filepath,multi_csv,"GetCurrent")
        return None


    def getInterpolatedPoints(self, tags: Union[str, list], timestamps: Union[list, int, float, str, datetime],
                              output: str = "json", filepath:str=None, multi_csv:bool=False):
        """
        Return a specified number of interpolated datapoints between a start and end timestamp for one or more datatags

        Args:
            tags: A datatag, or list of datatags
            timestamps: One or more timestamps for which to return a datapoint. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Timestamps, values and statuses for each datapoint returned, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args['tag'] = force_list(tags)
        args['timestamp'] = force_list(timestamps)
        response = self._doJsonRequest("getmulti", args)
        items = response['items']

        if response['unknown'] != []:
            if items == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(response["unknown"]) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(items)
        elif output == "df":
            return jsonToDf(items)
        elif output == "csv":
            return csvWriter(items,self.historian,filepath,multi_csv,"GetInterpolatedPoints")
        return None


    def getRawDataPoints(self, tags: Union[str, list], start: Union[int, float, str, datetime],
                         end: Union[int, float, str, datetime], maxpoints: int = 1000, output: str = "json", filepath:str=None, multi_csv:bool=False) -> object:
        """Return the first [maxpoints] raw datapoints between a start and end timestamp for one or more datatags, if maxpoints is greater than the total number of raw points, return all raw points within the time window

        Args:
            tags: A datatag, or list of datatags
            start: Start of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            end: End of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            maxpoints: (Optional) The maximum number of points to return. Defaults to 1000
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Timestamps, values and statuses for each datapoint returned, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args['tag'] = force_list(tags)
        args['start'] = time_to_epoch_millis(start)
        args['end'] = time_to_epoch_millis(end)
        args['maxpoints'] = maxpoints
        response = self._doJsonRequest("getraw", args)
        items = response["items"]

        if response['unknown'] != []:
            if items == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(response["unknown"]) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(items)
        elif output == "df":
            return jsonToDf(items)
        elif output == "csv":
            return csvWriter(items,self.historian,filepath,multi_csv,"GetRaw")
        return None



    def getInterpolatedRange(self, tags: Union[str, list], start: Union[int, float, str, datetime],
                             end: Union[int, float, str, datetime], count: int = 1000, output: str = "json", filepath:str=None, multi_csv:bool=False):
        """Return a specified number of interpolated datapoints between a start and end timestamp for one or more datatags

        Args:
            tags: A datatag, or list of datatags
            start: Start of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            end: End of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            count: (Optional) Number of points to return, evenly spaced across the given time range. Defaults to 1000
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Timestamps, values and statuses for each datapoint returned, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args['tag'] = force_list(tags)
        args['start'] = time_to_epoch_millis(start)
        args['end'] = time_to_epoch_millis(end)
        args['count'] = count
        response = self._doJsonRequest("getrange", args)
        items = response["items"]

        if response['unknown'] != []:
            if items == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(response["unknown"]) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(items)
        elif output == "df":
            return jsonToDf(items)
        elif output == "csv":
            return csvWriter(items,self.historian,filepath,multi_csv,"GetInterpolatedRange")
        return None

    def getAggregates(self, tags: Union[str, list], start: Union[int, float, str, datetime],
                      end: Union[int, float, str, datetime], fields: Union[str, list] = None, count: int = 1,
                      output: str = "json", filepath:str=None, multi_csv:bool=False):
        """
        Return specified aggregates of tags over a given time window, optionally divided into equal parts

        Args:
            tags: A datatag, or list of datatags
            start: Start of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            end: End of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            fields: (Optional) The aggregates to be calculated, accepts any subset of ["min","max","mean","stddev","var","numgood","numbad"]. Defaults to all aggregates.
            count: (Optional) Number of intervals to split the time window into, returning aggregates over each interval. Defaults to 1
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Timestamps, values and statuses for each datapoint returned, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        args = {}
        args['tag'] = force_list(tags)
        args['start'] = time_to_epoch_millis(start)
        args['end'] = time_to_epoch_millis(end)
        args['count'] = count
        args['aggfields'] = fields

        response = self._doJsonRequest("getagg", args)

        keys = set(response.keys())
        missing = [x for x in args['tag'] if x not in keys]

        if missing != []:
            if response == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(missing) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(response)
        elif output == "df":
            return aggToDf(response)
        elif output == "csv":
            order = ["tag","start","end","min","max","avg","var","stddev","count","numgood","numbad"]
            return csvWriter(response,self.historian,filepath,multi_csv,"GetAggregates",order,True)
        return None


    def getAggregateIntervals(self, tags, start: Union[int, float, str, datetime],
                              end: Union[int, float, str, datetime], window: str = None,
                              fields: Union[str, list] = None, output: str = "json", filepath:str=None, multi_csv:bool=False):
        """
        Return specified aggregates of tags over a given time window, optionally divided into equal parts

        Args:
            tags: A datatag, or list of datatags
            start: Start of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            end: End of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            window: (Optional) Sub-interval of time to divide the timeframe into as a multiple of seconds,minutes,hours,days or years. (e.g. "1d" for 1 day or "2h" for 2 hours etc). Defaults to the entire timeframe.
            fields: (Optional) The aggregates to be calculated, accepts any subset of ["min","max","mean","stddev","var","numgood","numbad"]. Defaults to all aggregates.
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            Timestamps, values and statuses for each datapoint returned, the format is dependent on the output parameter

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)

        epoch_ms_start = time_to_epoch_millis(start)
        epoch_ms_end = time_to_epoch_millis(end)

        if window is not None:
            windowDuration = parse_duration(window)
            totalDuration = epoch_ms_end - epoch_ms_start
            count = math.floor(totalDuration / windowDuration)
            epoch_ms_end = epoch_ms_end - totalDuration % windowDuration
        else:
            count = 1
        if count > 1000:
            logging.warning(
                "getAggregateIntervals is only designed to return up to 1000 aggregate sets, currently returning " + str(
                    count) + " sets. This may return overlapping intervals")

        args = {}
        args['tag'] = force_list(tags)
        args['start'] = epoch_ms_start
        args['end'] = epoch_ms_end
        args['count'] = count
        args['aggfields'] = fields
        response = self._doJsonRequest("getagg", args)

        keys = set(response.keys())
        missing = [x for x in args['tag'] if x not in keys]

        if missing != []:
            if response == {}:
                raise EigenException("None of the provided tags could be found in historian " + self.historian)
            logging.warning("Could not find tags: " + str(missing) + " in historian " + self.historian)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(response)
        elif output == "df":
            return aggToDf(response)
        elif output == "csv":
            order = ["tag","start","end","min","max","avg","var","stddev","count","numgood","numbad"]
            return csvWriter(response,self.historian,filepath,multi_csv,"GetAggregateIntervals",order,True)


    def countPoints(self, tags, start, end, output="json"):
        """
        Return the number of raw points of tags over a given time window

        Args:
            tags: A datatag, or list of datatags
            start: Start of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            end: End of the time range. Accepts epoch milliseconds, python datetime objects, or strings of format dd-mm-YYYYTHH:MM:DD
            output: (Optional) The format in which to return the data. Accepts one of: "raw" - The raw json returned by the API, "json" - A processed version of the json response, "df" - A formatted pandas dataframe object. Defaults to "json".

        Returns:
            The number of raw datapoints in the interval, per tag.

        Raises:
            ValueError: When invalid Output Type is given
            URLError: When no Ingenuity instance could be found at given EigenServer, or no internet connection is available
            RemoteServerException: When the provided historian could not be found, or the API response contains an error
            RunTimeError/HTTPError: No response from the API
        """
        validOutputTypes = ["raw", "json", "df", "csv"]
        if output not in validOutputTypes:
            raise ValueError("output must be one of %r." % validOutputTypes)
        ret = {}
        aggs = self.getAggregates(force_list(tags), start, end, "COUNT", 1, "raw")
        for tag in aggs.keys():
            ret[tag] = aggs[tag][0]["count"]

        response = self._matchArgumentCardinality(tags, ret)

        if output == "raw":
            return response
        elif output == "json":
            return json.dumps(response)
        elif output == "df":
            df = jsonToDf({"No. of Points": response})
            return df
        return None


    def _writeDataPoints(self, data:dict) -> bool:
        rawdata = {}
        for tag in list(data.keys()):
            rawdps = []
            for dp in force_list(data[tag]):
                rawdps.append([int(dp.getTimestamp() * 1000), dp.getValue(), dp.getStatus()])
            rawdata[tag] = rawdps

        return self._doJsonRequest("write", {'write': json.dumps(rawdata)})

    def createDataTag(self, tag:str, units:str, description:str) -> bool:
        """
        Create a new tag, or update the metadata of an existing tag

        Args:
            tag: The unique identifier of the tag
            units: The unit of the values stored in the tag
            description: A short explanation of the tags purpose or content

        Returns:
            A boolean representing the success of the creation. True means tag was created, False means the operation failed.
        """
        args = {}
        args['tag'] = tag
        args['units'] = units
        args['description'] = description
        return self._doJsonRequest("create", args)

    def _doJsonRequest(self, cmd, params):
        self._testConnection()
        url = self.serverurl + "?cmd=" + urlquote(cmd)
        return _do_eigen_json_request(url, **params)

    def _matchArgumentCardinality(self, proto, ret):
        """Takes in ret (a dict) and proto. If proto is a list, it returns
        ret. If proto is a single value, it extracts that key from ret and
        returns that instead.
        The intention is that:
            getTagThing("myTag") returns Thing for myTag
            getTagThing(["myTag", "myOtherTag"]) returns {'myTag': Thing, 'myOtherTag': OtherThing}
        and for clarification:
            getTagThing(["myTag"]) returns {'myTag': Thing}
        """
        try:
            badTags = []
            if is_list(proto):
                if ret != {}:
                    if len(ret) < len(proto):
                        for i in proto:
                            if i not in ret:
                                badTags.append(i)
                    return ret
                else:
                    raise KeyError
            return ret[proto]
        except KeyError:
            raise KeyError("Could not find tag(s): " + str(proto))


    def batchCSVImport(self, filepaths: Union[str,list]) -> bool:
        """
        Writes tag data from one or more csv file to Ingenuity. Will create a tag if it does not exist.

        Args:
            filepaths: The paths to one or more csv files. CSV Layouts must be as follows, with one point per line: tag,value,timestamp,status,units,description

        Returns:
            Success: A boolean representing the successful push of data. True means all csvs were successfully imported, False means at least one csv failed.
        """
        success = []
        for filepath in filepaths:
            with open(filepath, "r") as file:
                csv_reader = csv.reader(file, delimiter=',')
                data = list(csv_reader)
                # Upload in batches of 500 to avoid API timeouts/throttling
                chunks = divide_chunks(data, 500)
                for chunk in chunks:
                    datapoints = {}
                    for item in chunk:

                        self.createDataTag(item[0], item[4], item[5])
                        try:
                            datapoints[item[0]].append(DataPoint(item[1], int(item[2]), item[3]))
                        except KeyError:
                            datapoints[item[0]] = [DataPoint(item[1], int(item[2]), item[3])]
                    success.append(self._writeDataPoints(datapoints))
            file.close()
        return all(success)


    def batchJsonImport(self, data:Union[dict,list]) -> bool:
        """
        Writes datapoints from dict objects to tags in the Ingenuity historian.

        Args:
            data: a dict (or list of dicts) with format {"tag1": [{value,timestamp,status},{value,..},...], "tag2": {{value,..},...],...}

        Returns:
            A boolean indicating whether the push of data was successful. True means all data was successfully pushed, False means at least one batch failed to push.
        """
        success = []
        data = force_list(data)
        for datum in data:
            datapoints = {}
            totalPoints = 0
            tags = datum.keys()
            for tag in tags:
                # Upload in batches of 500 to avoid API timeouts/throttling
                if totalPoints >= 500:
                    self._writeDataPoints(datapoints)
                    datapoints = {}
                    totalPoints = 0

                for item in datum[tag]:
                    try:
                        datapoints[tag].append(DataPoint(item["value"],item["timestamp"],item["status"]))
                    except:
                        datapoints[tag] = [DataPoint(item["value"], item["timestamp"], item["status"])]
                totalPoints += len(datapoints[tag])

            success.append(self._writeDataPoints(datapoints))
        return all(success)


class DataPoint(object):
    """This class represents a data point which has a value and a python timestamp, and optionally a status of 'OK' or 'BAD'.
       Timestamp is stored in epochseconds. The constructor will convert datetime and tuples into floating point epoch seconds.
    """

    def __init__(self, value, timestamp, status = 'OK'):
        """This constructor takes in a value, a python timestamp (epoch floating point seconds or datetime or tuple)
        and an optional status.  Status is either "OK" or "BAD" and will default to "OK"
        """

        self.value = value

        if status is None:
            self.status = "OK"
        elif status == "OK" or self.status == "BAD":
            self.status = status
        else:
            raise EigenException("Unrecognised Status")

        # Convert timestamp from python datetime or tuple into python floating point epoch seconds
        self.datetime = datetime.fromtimestamp(timestamp)
        self.timestamp = pythonTimeToFloatingSecs(timestamp)


    def __str__(self):
        """Return a nicely formatted version of the datapoint"""
        val = self.value
        if type(val) != str:
            val = str(number_to_string(val))
        else:
            val = val.ljust(12)
        return val + " @ " + get_timestamp_string(self.timestamp) + " - " + self.status

    def __repr__(self):
        return "DataPoint[" + str(self) + "]"

    def getValue(self):
        """Return the value of the datapoint"""
        return self.value

    def getTimestamp(self):
        """Return the timestamp of the datapoint in python floating point epoch seconds"""
        return self.timestamp

    def getTimestampMillis(self):
        """Return the timestamp of the datapoint in epochmillis"""
        return pythonTimeToServerTime(self.timestamp)

    def getTimestampAsDatetime(self):
        """Return the timestamp of the datapoint as Datetime"""
        return get_datetime(self.timestamp)

    def getTimestampAsTuple(self):
        """Return the timestamp of the datapoint as Tuple"""
        return get_time_tuple(self.timestamp)

    def getStatus(self):
        """Return the status of the datapoint - either 'OK' or 'BAD'"""
        return self.status

    def isBad(self):
        """Returns either True if status = "BAD" or false if status = "OK" """
        return self.status != "OK"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def getAsJson(self):
        """Return as a json map with the timestamp in milliseconds"""
        return "{" + "\"value\":" + str(self.value) + ",\"timestamp\":" + self.getTimestampMillis(
            self) + ",\"status\":\"" + self.status + "\"}"


def get_datapoint(dp):
    tag = dp["tag"]
    value = dp["value"]
    timestamp = dp["timestamp"]
    timestampinmillis = dp["timestamp"]
    timestamp = serverTimeToPythonTime(timestampinmillis)
    status = dp["status"]
    return DataPoint(tag, value, timestamp, status)


def get_historian(instance: str = None, eigenserver: EigenServer = None) -> JsonBridgeHistorian:
    """Instantiate a historian object for the given instance.

    Args:
        instance: (Optional) The name of the historian to connect to. Defaults to the default historian of the EigenServer.
        eigenserver: (Optional) An EigenServer Object linked to the ingenuity url containing the historian. Can be omitted if environmental variable "EIGENSERVER" exists and is equal to the Ingenuity base url

    Returns:
        An Object that can be used to query historian data.

    """
    if eigenserver is None:
        eigenserver = get_default_server()

    if instance is not None and (instance.startswith("http:") or instance.startswith("https:")):
        return JsonBridgeHistorian(instance)

    if instance is None:
        instance = get_default_historian_name(eigenserver=eigenserver)
        if instance is None:
            raise EigenException("No default historian instance found")
    return JsonBridgeHistorian(eigenserver.getEigenServerUrl() + "historian-servlet" + "/jsonbridge/" + instance)


def list_historians(eigenserver=None) -> list:
    """List all historian instances in EigenServer.

    Args:
        eigenserver: (Optional) An EigenServer Object linked to the ingenuity url containing the historian. Can be omitted if environmental variable "EIGENSERVER" exists and is equal to the Ingenuity base url

    Returns:
        A list of all historians on the server
    """
    if eigenserver is None:
        eigenserver = get_default_server()

    return eigenserver.listDataSources("historian")


def get_default_historian_name(eigenserver=None) -> str:
    """Get the default historian of an EigenServer.

    Args:
        eigenserver: (Optional) An EigenServer Object linked to the ingenuity url containing the historian. Can be omitted if environmental variable "EIGENSERVER" exists and is equal to the Ingenuity base url

    Returns:
        The name of the default historian
    """
    if eigenserver is None:
        eigenserver = get_default_server()

    return eigenserver.getDefaultDataSource("historian")
